from __future__ import annotations

from collections import Counter, defaultdict
from pathlib import Path
from typing import Generator

import midsv

from DAJIN2.utils import cssplits_handler, io, sam_handler


def has_inversion_in_splice(CIGAR: str) -> bool:
    previous_insertion = False
    for cigar in sam_handler.split_cigar(CIGAR):
        if cigar.endswith("I"):
            previous_insertion = True
            continue
        if previous_insertion and cigar.endswith("N"):
            return True
        else:
            previous_insertion = False
    return False


def extract_preset_and_cigar_by_qname(path_sam_files: list[Path]) -> dict[dict[str, str]]:
    preset_cigar_by_qname = defaultdict(dict)
    # Extract preset and CIGAR
    for path in path_sam_files:
        preset = path.stem
        sam: list[list[str]] = io.read_sam(path)
        for record in sam:
            if record[0].startswith("@"):
                continue
            qname = record[0]
            cigar = record[5]
            if qname in preset_cigar_by_qname and preset in preset_cigar_by_qname[qname]:
                cigar += preset_cigar_by_qname[qname][preset]
            preset_cigar_by_qname[qname].update({preset: cigar})
    return dict(preset_cigar_by_qname)


def extract_best_preset(preset_cigar_by_qname: dict[str, dict[str, str]]) -> dict[str, str]:
    best_preset = defaultdict(str)
    for qname in preset_cigar_by_qname:
        preset_cigar = preset_cigar_by_qname[qname]
        # If there is an inversion in the splice preset, remove it
        if "splice" in preset_cigar and has_inversion_in_splice(preset_cigar["splice"]):
            preset_cigar.pop("splice")

        if preset_cigar == {}:
            continue

        alignment_lengths = {
            preset: sam_handler.calculate_alignment_length(cigar) for preset, cigar in preset_cigar.items()
        }

        # If all alignment lengths are the same, prioritize map-ont
        if len(set(alignment_lengths.values())) == 1 and "map-ont" in alignment_lengths:
            best_preset[qname] = "map-ont"
            continue

        # Define a custom key function to prioritize map-ont
        def custom_key(key: str, alignment_lengths=alignment_lengths) -> tuple[int, bool]:
            return (alignment_lengths[key], key == "map-ont")

        max_key = max(alignment_lengths, key=custom_key)
        best_preset[qname] = max_key

    return dict(best_preset)


def extract_best_alignment_length_from_sam(
    path_sam_files: list[Path], best_preset: dict[str, str]
) -> Generator[list[str]]:
    flag_header = False
    for path in path_sam_files:
        preset = path.stem
        sam = io.read_sam(path)
        for record in sam:
            if record[0].startswith("@"):
                if not flag_header:
                    yield record
            else:
                qname = record[0]
                if best_preset.get(qname) == preset:
                    yield record
        flag_header = True


def transform_to_midsv_format(sam: Generator[list[str]]) -> Generator[list[dict]]:
    yield from midsv.transform(sam, midsv=False, cssplit=True, qscore=False, keep={"FLAG"})


def replace_internal_n_to_d(midsv_sample: Generator[list[dict]], sequence: str) -> Generator[list[dict]]:
    """
    Replace internal 'N's with 'D' in a given sequence.
    This function modifies the 'CSSPLIT' field in the input sample. It identifies
    the boundaries of consecutive 'N's and replaces any 'N' within these boundaries
    with the corresponding character from the provided sequence. 'N's at the boundaries
    remain unchanged.
    """
    for samp in midsv_sample:
        cssplits = samp["CSSPLIT"].split(",")

        left_idx_n, right_idx_n = cssplits_handler.find_n_boundaries(cssplits)

        for j, (cs, seq_char) in enumerate(zip(cssplits, sequence)):
            if left_idx_n < j < right_idx_n and cs == "N":
                cssplits[j] = f"-{seq_char}"

        samp["CSSPLIT"] = ",".join(cssplits)
        yield samp


def is_reverse_strand(flag: int) -> bool:
    """
    Determines if the read is mapped to the reverse strand.
    """
    # Check if the bit 4 (0-based) is set
    return bool(flag & 16)


def convert_flag_to_strand(midsv_sample: Generator[list[dict]]) -> Generator[list[dict]]:
    """Convert FLAG to STRAND (+ or -)"""
    for samp in midsv_sample:
        samp["STRAND"] = "-" if is_reverse_strand(int(samp["FLAG"])) else "+"
        del samp["FLAG"]
        yield samp


def filter_samples_by_n_proportion(midsv_sample: Generator[dict], threshold: int = 95) -> Generator[list[dict]]:
    """Filters out the samples from the input generator where the proportion of 'N' in the 'CSSPLIT' field is 95% or higher."""
    for samp in midsv_sample:
        cssplits = samp.get("CSSPLIT", "").split(",")
        count = Counter(cssplits)
        n_percentage = count["N"] / sum(count.values()) * 100
        if n_percentage < threshold:
            yield samp


###########################################################
# convert_consecutive_indels_to_match
###########################################################


def convert_consecutive_indels_to_match(cssplit: str) -> str:
    i = 0
    cssplit_reversed = cssplit.split(",")[::-1]
    while i < len(cssplit_reversed):
        current_cs = cssplit_reversed[i]

        if not current_cs.startswith("+"):
            i += 1
            continue

        insertions = [base.lstrip("+") for base in current_cs.split("|")[:-1]][::-1]

        # Extract deletions
        deletions = []

        for j in range(1, len(insertions) + 1):
            if i + j >= len(cssplit_reversed):
                break

            next_cs = cssplit_reversed[i + j]

            if not next_cs.startswith("-"):
                break

            deletions.append(next_cs.lstrip("-"))

        if insertions != deletions:
            i += 1
            continue

        # Format insertions
        cssplit_reversed[i] = current_cs.split("|")[-1]

        # Format deletions
        for k, _ in enumerate(insertions, 1):
            cssplit_reversed[i + k] = cssplit_reversed[i + k].replace("-", "=")

        i += len(insertions) + 1

    return ",".join(cssplit_reversed[::-1])


def convert_consecutive_indels(midsv_sample: Generator) -> Generator[list[dict]]:
    """
    Due to alignment errors, there can be instances where a true match is mistakenly replaced with "insertion following a deletion".
    For example, although it should be "=C,=T", it gets replaced by "-C,+C|=T". In such cases, a process is performed to revert it back to "=C,=T".
    """
    for m in midsv_sample:
        m["CSSPLIT"] = convert_consecutive_indels_to_match(m["CSSPLIT"])
        yield m


###########################################################
# main
###########################################################


def generate_midsv(ARGS, is_control: bool = False, is_sv: bool = False) -> None:
    name = ARGS.control_name if is_control else ARGS.sample_name

    for allele, sequence in ARGS.fasta_alleles.items():
        path_midsv_directory = Path(ARGS.tempdir, name, "midsv", allele)
        path_midsv_directory.mkdir(parents=True, exist_ok=True)

        if Path(path_midsv_directory, f"{name}.jsonl").exists():
            continue

        if is_control and is_sv:
            """
            Set the destination for midsv as `barcode01/midsv/insertion1_barcode02.json` when control is barcode01, sample is barcode02, and the allele is insertion1.
            """
            path_sam_files = list(Path(ARGS.tempdir, name, "sam", allele).glob(f"{ARGS.sample_name}_*.sam"))
            path_midsv_output = Path(ARGS.tempdir, name, "midsv", allele, f"{ARGS.sample_name}.jsonl")
        else:
            """
            Set the destination for midsv as `barcode02/midsv/insertion1.json` when the sample is barcode02 and the allele is insertion1.
            """
            path_sam_files = list(Path(ARGS.tempdir, name, "sam", allele).glob("*.sam"))
            path_midsv_output = Path(ARGS.tempdir, name, "midsv", allele, f"{name}.jsonl")

        preset_cigar_by_qname = extract_preset_and_cigar_by_qname(path_sam_files)
        best_preset = extract_best_preset(preset_cigar_by_qname)
        sam_best_alignments = extract_best_alignment_length_from_sam(path_sam_files, best_preset)
        midsv_chaind = transform_to_midsv_format(sam_best_alignments)
        midsv_sample = replace_internal_n_to_d(midsv_chaind, sequence)
        midsv_sample = convert_flag_to_strand(midsv_sample)
        midsv_sample = filter_samples_by_n_proportion(midsv_sample)
        midsv_sample = convert_consecutive_indels(midsv_sample)
        midsv.write_jsonl(midsv_sample, path_midsv_output)



from __future__ import annotations

import shutil
from pathlib import Path

import plotly.express as px
from DAJIN2.utils import io
from DAJIN2.utils.config import DAJIN_RESULTS_DIR, TEMP_ROOT_DIR


def rename_allele(allele: str) -> str:
    if allele == "control" or allele.startswith("insertion"):
        return allele.capitalize()
    return allele


def rename_type(type_: str) -> str:
    if type_ == "intact" or type_ == "indels":
        return type_.capitalize()
    elif type_ == "sv":
        return "SV"
    return type_


def format_result_info(path_result: Path) -> list[dict[str, str]]:
    sample_name = path_result.stem

    key_filter = ["QNAME", "NAME", "READNUM", "PERCENT"]
    key_mapping = {"QNAME": "Read ID", "READNUM": "Number of reads", "PERCENT": "Percent of reads"}
    key_order = ["Sample", "Read ID", "Label", "Allele", "Type", "Number of reads", "Percent of reads"]

    result_format = []
    for reads in io.read_jsonl(path_result):
        # Filter keys
        reads = {k: reads[k] for k in key_filter}
        # Add Label, Allele, Type from NAME
        label, allele, type_, *_ = reads["NAME"].split("_")
        reads.update({"Label": label.capitalize(), "Allele": rename_allele(allele), "Type": rename_type(type_)})
        del reads["NAME"]
        # Add Sample
        reads["Sample"] = sample_name
        # Rename keys
        for old_key, new_key in key_mapping.items():
            reads[new_key] = reads.pop(old_key)
        # Reorder keys
        reads_order = {k: reads[k] for k in key_order}
        result_format.append(reads_order)

    return result_format


def extract_all_info(batch_directory: Path) -> list[dict[str, str]]:
    results_all = []
    for path_result in sorted(batch_directory.iterdir()):
        result_format = format_result_info(path_result)
        results_all.extend(result_format)
    return results_all


def summarize_info(results_all: list[dict[str, str]]) -> list[dict[str, str]]:
    # Delete the ID column and extract unique rows
    unique_rows = []
    seen = set()  # A set to record the rows that have already been seen

    for row in results_all:
        # Delete the Read ID key
        row.pop("Read ID", None)
        # Convert to a tuple format to be able to add to the set, and add only unique rows to the seen set
        row_tuple = tuple(row.items())
        if row_tuple not in seen:
            seen.add(row_tuple)
            unique_rows.append(row)

    return sorted(unique_rows, key=lambda x: [x["Sample"], x["Label"]])


def add_allele_type(results_summary: list[dict[str, str]]) -> list[dict[str, str]]:
    for row in results_summary:
        row["Allele type"] = f"{row['Allele']} {row['Type']}"
    return results_summary


def order_allele_type(results_summary: list[dict[str, str]]) -> list[str]:
    alleles = {a["Allele"] for a in results_summary}
    alleles_insertion = {a for a in alleles if a.startswith("Insertion")}
    alleles_order = ["Control"] + sorted(alleles - alleles_insertion - {"Control"}) + sorted(alleles_insertion)

    allele_type_order = []
    for allele in alleles_order:
        for type_ in ["Intact", "Indels", "SV"]:
            allele_type_order.append(f"{allele} {type_}")

    allele_type = {a["Allele type"] for a in results_summary}
    return [at for at in allele_type_order if at in allele_type]


def output_plot(results_summary: list[dict[str, str]], report_directory: Path):
    results_plot = add_allele_type(results_summary)
    alleletype_order = order_allele_type(results_summary)

    fig = px.bar(
        results_plot,
        x="Sample",
        y="Percent of reads",
        color="Allele type",
        text="Percent of reads",
        labels={"Sample": "Samples", "Percent of reads": "Percent of reads", "Allele type": "Alelle type"},
        category_orders={"Allele type": alleletype_order},
    )
    fig.update_traces(textposition="inside", cliponaxis=False)
    fig.update_xaxes(categoryorder="category ascending")

    output_filename = Path(report_directory, "read_plot")
    fig.write_html(f"{output_filename}.html")
    # if kaleido is installed, output a pdf
    try:
        fig.write_image(f"{output_filename}.pdf")
    except Exception:
        pass


def report(NAME: str) -> None:
    report_directory = Path(DAJIN_RESULTS_DIR, NAME)
    report_directory.mkdir(exist_ok=True, parents=True)
    shutil.copytree(Path(TEMP_ROOT_DIR, NAME, "report"), report_directory, dirs_exist_ok=True)

    results_all = extract_all_info(Path(TEMP_ROOT_DIR, NAME, "result"))
    results_summary = summarize_info(results_all)

    # Write to Excel
    io.write_xlsx(results_summary, Path(report_directory, "read_summary.xlsx"))

    # Write to plot as HTML and PDF
    output_plot(results_summary, report_directory)
